@@ -1,5 +1,5 @@
 #
-# Copyright 2017-2018 European Centre for Medium-Range Weather Forecasts (ECMWF).
+# Copyright 2017-2019 European Centre for Medium-Range Weather Forecasts (ECMWF).
 #
 # Licensed under the Apache License, Version 2.0 (the "License");
 # you may not use this file except in compliance with the License.
@@ -17,10 +17,9 @@
 #   Alessandro Amici - B-Open - https://bopen.eu
 #
 
-from __future__ import absolute_import, division, print_function, unicode_literals
-
 import collections
 import functools
+import logging
 import typing as T  # noqa
 
 import xarray as xr  # noqa
@@ -30,6 +29,7 @@ from . import cfunits
 
 COORD_MODEL = {}  # type: T.Dict[str, T.Dict[str, T.Any]]
 COORD_TRANSLATORS = collections.OrderedDict()  # type: T.Dict[str, T.Callable]
+LOG = logging.getLogger(__name__)
 
 
 def match_values(match_value_func, mapping):
@@ -55,8 +55,13 @@ def translate_direction(data, out_name, stored_direction):
 
 
 def coord_translator(
-        default_out_name, default_units, default_direction, is_cf_type, cf_type, data,
-        coord_model=COORD_MODEL
+    default_out_name,
+    default_units,
+    default_direction,
+    is_cf_type,
+    cf_type,
+    data,
+    coord_model=COORD_MODEL,
 ):
     # type: (str, str, str, T.Callable, str, xr.DataArray, dict) -> xr.DataArray
     out_name = coord_model.get(cf_type, {}).get('out_name', default_out_name)
@@ -91,7 +96,7 @@ def is_latitude(coord):
 
 
 COORD_TRANSLATORS['latitude'] = functools.partial(
-    coord_translator, 'latitude', 'degrees_north', 'decreasing', is_latitude,
+    coord_translator, 'latitude', 'degrees_north', 'decreasing', is_latitude
 )
 
 
@@ -104,7 +109,7 @@ def is_longitude(coord):
 
 
 COORD_TRANSLATORS['longitude'] = functools.partial(
-    coord_translator, 'longitude', 'degrees_east', 'increasing', is_longitude,
+    coord_translator, 'longitude', 'degrees_east', 'increasing', is_longitude
 )
 
 
@@ -117,7 +122,7 @@ TIME_CF_UNITS = 'seconds since 1970-01-01T00:00:00+00:00'
 
 
 COORD_TRANSLATORS['time'] = functools.partial(
-    coord_translator, 'time', TIME_CF_UNITS, 'increasing', is_time,
+    coord_translator, 'time', TIME_CF_UNITS, 'increasing', is_time
 )
 
 
@@ -126,9 +131,7 @@ def is_step(coord):
     return coord.attrs.get('standard_name') == 'forecast_period'
 
 
-COORD_TRANSLATORS['step'] = functools.partial(
-    coord_translator, 'step', 'h', 'increasing', is_step,
-)
+COORD_TRANSLATORS['step'] = functools.partial(coord_translator, 'step', 'h', 'increasing', is_step)
 
 
 def is_valid_time(coord):
@@ -141,7 +144,17 @@ def is_valid_time(coord):
 
 
 COORD_TRANSLATORS['valid_time'] = functools.partial(
-    coord_translator, 'valid_time', TIME_CF_UNITS, 'increasing', is_valid_time,
+    coord_translator, 'valid_time', TIME_CF_UNITS, 'increasing', is_valid_time
+)
+
+
+def is_depth(coord):
+    # type: (xr.Coordinate) -> bool
+    return coord.attrs.get('standard_name') == 'depth'
+
+
+COORD_TRANSLATORS['depthBelowLand'] = functools.partial(
+    coord_translator, 'depthBelowLand', 'm', 'decreasing', is_depth
 )
 
 
@@ -151,7 +164,7 @@ def is_isobaric(coord):
 
 
 COORD_TRANSLATORS['isobaricInhPa'] = functools.partial(
-    coord_translator, 'isobaricInhPa', 'hPa', 'decreasing', is_isobaric,
+    coord_translator, 'isobaricInhPa', 'hPa', 'decreasing', is_isobaric
 )
 
 
@@ -161,20 +174,35 @@ def is_number(coord):
 
 
 COORD_TRANSLATORS['number'] = functools.partial(
-    coord_translator, 'number', '1', 'increasing', is_number,
+    coord_translator, 'number', '1', 'increasing', is_number
 )
 
 
 def translate_coords(
-        data, coord_model=COORD_MODEL, errors='strict', coord_translators=COORD_TRANSLATORS
+    data, coord_model=COORD_MODEL, errors='warn', coord_translators=COORD_TRANSLATORS
 ):
     # type: (xr.Dataset, T.Dict, str, T.Dict) -> xr.Dataset
     for cf_name, translator in coord_translators.items():
         try:
             data = translator(cf_name, data, coord_model=coord_model)
         except:
-            if errors != 'ignore':
+            if errors == 'ignore':
+                pass
+            elif errors == 'raise':
                 raise RuntimeError("error while translating coordinate: %r" % cf_name)
+            else:
+                LOG.warning("error while translating coordinate: %r", cf_name)
+    config = coord_model.get('config', {})
+    if config.get('preferred_time_dimension', 'time') == 'valid_time':
+        try:
+            data = ensure_valid_time(data)
+        except Exception:
+            if errors == 'ignore':
+                pass
+            elif errors == 'raise':
+                raise RuntimeError("error while ensuring valid_time coordinate")
+            else:
+                LOG.exception("error while ensuring valid_time coordinate")
     return data
 
 
@@ -183,14 +211,13 @@ def ensure_valid_time_present(data, valid_time_name='valid_time'):
     valid_times = match_values(is_valid_time, data.coords)
     times = match_values(is_time, data.coords)
     steps = match_values(is_step, data.coords)
-    time = step = ''
+    time = times[0] if times else ''
+    step = steps[0] if steps else ''
     if not valid_times:
-        if not times:
+        if not time:
             raise ValueError("not enough information to ensure a 'valid_time'.")
         valid_time = valid_time_name
-        time = times[0]
-        if steps:
-            step = steps[0]
+        if step:
             data.coords[valid_time] = data.coords[time] + data.coords[step]
         else:
             data.coords[valid_time] = data.coords[time]
@@ -204,8 +231,18 @@ def ensure_valid_time(data):
     # type: (xr.Dataset) -> xr.Dataset
     valid_time, time, step = ensure_valid_time_present(data)
     if valid_time not in data.dims:
-        if data.coords[time].size == data.coords[valid_time].size:
+        if time and time in data.dims and data.coords[time].size == data.coords[valid_time].size:
             return data.swap_dims({time: valid_time})
-        if data.coords[step].size == data.coords[valid_time].size:
+        if step and step in data.dims and data.coords[step].size == data.coords[valid_time].size:
             return data.swap_dims({step: valid_time})
+        # also convert is valid_time can index all times and steps
+        if (
+            step
+            and time
+            and step in data.dims
+            and time in data.dims
+            and data.coords[step].size * data.coords[time].size == data.coords[valid_time].size
+        ):
+            data = data.stack(tmp_coord=(time, step))
+            data = data.swap_dims({'tmp_coord': valid_time}).drop('tmp_coord').dropna(valid_time)
     return data
