@@ -1,5 +1,5 @@
 #
-# Copyright 2017-2018 European Centre for Medium-Range Weather Forecasts (ECMWF).
+# Copyright 2017-2019 European Centre for Medium-Range Weather Forecasts (ECMWF).
 #
 # Licensed under the Apache License, Version 2.0 (the "License");
 # you may not use this file except in compliance with the License.
@@ -17,15 +17,6 @@
 #   Alessandro Amici - B-Open - https://bopen.eu
 #
 
-from __future__ import absolute_import, division, print_function, unicode_literals
-from builtins import bytes, isinstance, str, type
-
-# Python 2 compatibility bit not in python-future
-try:
-    FileExistsError
-except NameError:
-    FileExistsError = OSError
-
 import collections
 import contextlib
 import hashlib
@@ -37,81 +28,99 @@ import typing as T
 
 import attr
 
-from . import eccodes
+# select between using the external ecCodes bindings or the internal implementation
+if os.environ.get('CFGRIB_USE_EXTERNAL_ECCODES_BINDINGS'):
+    import eccodes
+else:
+    from . import bindings as eccodes
 
+eccodes_version = eccodes.codes_get_api_version()
 
 LOG = logging.getLogger(__name__)
 _MARKER = object()
 
+#
+# No explicit support for MULTI-FIELD at Message level.
+#
+eccodes.codes_grib_multi_support_on()
+
 
 @attr.attrs()
-class Message(collections.MutableMapping):
+class Message(collections.abc.MutableMapping):
     """Dictionary-line interface to access Message headers."""
+
     codes_id = attr.attrib()
     encoding = attr.attrib(default='ascii', type=str)
-    errors = attr.attrib(default='ignore', validator=attr.validators.in_(['ignore', 'strict']))
+    errors = attr.attrib(
+        default='warn', validator=attr.validators.in_(['ignore', 'warn', 'raise'])
+    )
 
     @classmethod
-    def from_file(cls, file, offset=None, product_kind=eccodes.CODES_PRODUCT_GRIB, **kwargs):
-        # type: (T.IO[bytes], int, int, T.Any) -> Message
+    def from_file(cls, file, offset=None, **kwargs):
+        # type: (T.IO[bytes], int, T.Any) -> Message
+        field_in_message = 0
+        if isinstance(offset, tuple):
+            offset, field_in_message = offset
         if offset is not None:
             file.seek(offset)
-        codes_id = eccodes.codes_handle_new_from_file(file, product_kind)
+        codes_id = None
+        # iterate over multi-fields in the message
+        for _ in range(field_in_message + 1):
+            codes_id = eccodes.codes_grib_new_from_file(file)
+        if codes_id is None:
+            raise EOFError("End of file: %r" % file)
         return cls(codes_id=codes_id, **kwargs)
 
     @classmethod
     def from_sample_name(cls, sample_name, product_kind=eccodes.CODES_PRODUCT_GRIB, **kwargs):
-        codes_id = eccodes.codes_new_from_samples(sample_name.encode('ASCII'), product_kind)
+        codes_id = eccodes.codes_new_from_samples(sample_name, product_kind)
         return cls(codes_id=codes_id, **kwargs)
 
     @classmethod
     def from_message(cls, message, **kwargs):
-        codes_id = eccodes.codes_handle_clone(message.codes_id)
+        codes_id = eccodes.codes_clone(message.codes_id)
         return cls(codes_id=codes_id, **kwargs)
 
     def __del__(self):
-        eccodes.codes_handle_delete(self.codes_id)
+        eccodes.codes_release(self.codes_id)
 
-    def message_get(self, item, key_type=None, size=None, length=None, default=_MARKER):
-        # type: (str, int, int, int, T.Any) -> T.Any
+    def message_get(self, item, key_type=None, default=_MARKER):
+        # type: (str, type, T.Any) -> T.Any
         """Get value of a given key as its native or specified type."""
-        key = item.encode(self.encoding)
+        key = item
         try:
-            values = eccodes.codes_get_array(self.codes_id, key, key_type, size, length)
-        except eccodes.EcCodesError as ex:
-            if ex.code == eccodes.lib.GRIB_NOT_FOUND:
-                if default is _MARKER:
-                    raise KeyError(item)
-                else:
-                    return default
-            else:  # pragma: no cover
-                raise
-        if values and isinstance(values[0], bytes):
-            values = [v.decode(self.encoding) for v in values]
+            values = eccodes.codes_get_array(self.codes_id, key, key_type)
+            if values is None:
+                values = ['unsupported_key_type']
+        except eccodes.KeyValueNotFoundError:
+            if default is _MARKER:
+                raise KeyError(item)
+            else:
+                return default
         if len(values) == 1:
             return values[0]
         return values
 
     def message_set(self, item, value):
         # type: (str, T.Any) -> None
-        key = item.encode(self.encoding)
+        key = item
         set_array = isinstance(value, T.Sequence) and not isinstance(value, (str, bytes))
         if set_array:
             eccodes.codes_set_array(self.codes_id, key, value)
         else:
             if isinstance(value, str):
-                value = value.encode(self.encoding)
+                value = value
             eccodes.codes_set(self.codes_id, key, value)
 
     def message_iterkeys(self, namespace=None):
         # type: (str) -> T.Generator[str, None, None]
         if namespace is not None:
-            bnamespace = namespace.encode(self.encoding)  # type: T.Optional[bytes]
+            bnamespace = namespace  # type: T.Optional[str]
         else:
             bnamespace = None
         iterator = eccodes.codes_keys_iterator_new(self.codes_id, namespace=bnamespace)
         while eccodes.codes_keys_iterator_next(iterator):
-            yield eccodes.codes_keys_iterator_get_name(iterator).decode(self.encoding)
+            yield eccodes.codes_keys_iterator_get_name(iterator)
         eccodes.codes_keys_iterator_delete(iterator)
 
     def __getitem__(self, item):
@@ -122,12 +131,17 @@ class Message(collections.MutableMapping):
         # type: (str, T.Any) -> None
         try:
             return self.message_set(item, value)
-        except eccodes.EcCodesError as ex:
-            if self.errors == 'ignore' and ex.code == eccodes.lib.GRIB_READ_ONLY:
-                # Very noisy error when trying to set computed keys
+        except eccodes.GribInternalError as ex:
+            if self.errors == 'ignore':
                 pass
-            else:
+            elif self.errors == 'raise':
                 raise KeyError("failed to set key %r to %r" % (item, value))
+            else:
+                if isinstance(ex, eccodes.ReadOnlyError):
+                    # Very noisy error when trying to set computed keys
+                    pass
+                else:
+                    LOG.warning("failed to set key %r to %r", item, value)
 
     def __delitem__(self, item):
         raise NotImplementedError
@@ -148,6 +162,7 @@ class Message(collections.MutableMapping):
 @attr.attrs()
 class ComputedKeysMessage(Message):
     """Extension of Message class for adding computed keys."""
+
     computed_keys = attr.attrib(
         default={},
         type=T.Dict[str, T.Tuple[T.Callable[[Message], T.Any], T.Callable[[Message], T.Any]]],
@@ -177,32 +192,15 @@ class ComputedKeysMessage(Message):
             return super(ComputedKeysMessage, self).__setitem__(item, value)
 
 
-def make_message_schema(message, schema_keys, log=LOG):
-    schema = collections.OrderedDict()
-    for key in schema_keys:
-        bkey = key.encode(message.encoding)
-        try:
-            key_type = eccodes.codes_get_native_type(message.codes_id, bkey)
-        except eccodes.EcCodesError as ex:
-            if ex.code != eccodes.lib.GRIB_NOT_FOUND:  # pragma: no cover
-                log.exception("key %r failed", key)
-            schema[key] = ()
-            continue
-        size = eccodes.codes_get_size(message.codes_id, bkey)
-        if key_type == eccodes.CODES_TYPE_STRING:
-            length = eccodes.codes_get_length(message.codes_id, bkey)
-            schema[key] = (key_type, size, length)
-        else:
-            schema[key] = (key_type, size)
-    return schema
-
-
 @attr.attrs()
-class FileStream(collections.Iterable):
+class FileStream(collections.abc.Iterable):
     """Iterator-like access to a filestream of Messages."""
-    path = attr.attrib(type=str)
+
+    path = attr.attrib(type=str, converter=os.path.abspath)
     message_class = attr.attrib(default=Message, type=Message, repr=False)
-    errors = attr.attrib(default='ignore', validator=attr.validators.in_(['ignore', 'strict']))
+    errors = attr.attrib(
+        default='warn', validator=attr.validators.in_(['ignore', 'warn', 'raise'])
+    )
 
     def __iter__(self):
         # type: () -> T.Generator[Message, None, None]
@@ -218,9 +216,11 @@ class FileStream(collections.Iterable):
                     break
                 except Exception:
                     if self.errors == 'ignore':
-                        LOG.exception("skipping corrupted Message")
-                    else:
+                        pass
+                    elif self.errors == 'raise':
                         raise
+                    else:
+                        LOG.exception("skipping corrupted Message")
 
     def message_from_file(self, file, offset=None, **kwargs):
         return self.message_class.from_file(file=file, offset=offset, **kwargs)
@@ -241,40 +241,45 @@ def compat_create_exclusive(path, *args, **kwargs):
         try:
             yield file
         except Exception:
+            file.close()
             os.unlink(path)
             raise
 
 
 @attr.attrs()
-class FileIndex(collections.Mapping):
+class FileIndex(collections.abc.Mapping):
+    allowed_protocol_version = '1'
     filestream = attr.attrib(type=FileStream)
     index_keys = attr.attrib(type=T.List[str])
     offsets = attr.attrib(repr=False, type=T.List[T.Tuple[T.Tuple[T.Any, ...], T.List[int]]])
+    filter_by_keys = attr.attrib(default={}, type=T.Dict[str, T.Any])
 
     @classmethod
     def from_filestream(cls, filestream, index_keys):
-        # FIXME: using `Message.message_get` with an explicit message schema was a significant
-        #   optimization at some point, due to less calls to the slow CFFI ABI interface.
-        #   This doesn't appear to be reproducible at the moment so the optimisation is
-        #   disabled and we may choose to remove `make_message_schema` altogether.
-        schema = make_message_schema(filestream.first(), index_keys)
         offsets = collections.OrderedDict()
+        count_offsets = {}  # type: T.Dict[int, int]
         for message in filestream:
             header_values = []
-            for key, args in schema.items():
+            for key in index_keys:
                 try:
-                    # if args and not key == 'time':
-                    #     value = message.message_get(key, *args)
-                    # else:
                     value = message[key]
                 except:
                     value = 'undef'
                 if isinstance(value, list):
                     value = tuple(value)
                 header_values.append(value)
-            offset = message.message_get('offset', eccodes.CODES_TYPE_LONG)
-            offsets.setdefault(tuple(header_values), []).append(offset)
-        return cls(filestream=filestream, index_keys=index_keys, offsets=list(offsets.items()))
+            offset = message.message_get('offset', int)
+            if offset in count_offsets:
+                count_offsets[offset] += 1
+                offset_field = (offset, count_offsets[offset])
+            else:
+                count_offsets[offset] = 0
+                offset_field = offset
+            offsets.setdefault(tuple(header_values), []).append(offset_field)
+        self = cls(filestream=filestream, index_keys=index_keys, offsets=list(offsets.items()))
+        # record the index protocol version in the instance so it is dumped with pickle
+        self.index_protocol_version = cls.allowed_protocol_version
+        return self
 
     @classmethod
     def from_indexpath(cls, indexpath):
@@ -283,9 +288,14 @@ class FileIndex(collections.Mapping):
 
     @classmethod
     def from_indexpath_or_filestream(
-            cls, filestream, index_keys, indexpath='{path}.{short_hash}.idx', log=LOG,
+        cls, filestream, index_keys, indexpath='{path}.{short_hash}.idx', log=LOG
     ):
         # type: (FileStream, T.List[str], str, logging.Logger) -> FileIndex
+
+        # Reading and writing the index can be explicitly suppressed by passing indexpath==''.
+        if not indexpath:
+            return cls.from_filestream(filestream, index_keys)
+
         hash = hashlib.md5(repr(index_keys).encode('utf-8')).hexdigest()
         indexpath = indexpath.format(path=filestream.path, hash=hash, short_hash=hash[:5])
         try:
@@ -303,8 +313,12 @@ class FileIndex(collections.Mapping):
             filestream_mtime = os.path.getmtime(filestream.path)
             if index_mtime >= filestream_mtime:
                 self = cls.from_indexpath(indexpath)
-                if getattr(self, 'index_keys', None) == index_keys and \
-                        getattr(self, 'filestream', None) == filestream:
+                allowed_protocol_version = self.allowed_protocol_version
+                if (
+                    getattr(self, 'index_keys', None) == index_keys
+                    and getattr(self, 'filestream', None) == filestream
+                    and getattr(self, 'index_protocol_version', None) == allowed_protocol_version
+                ):
                     return self
                 else:
                     log.warning("Ignoring index file %r incompatible with GRIB file", indexpath)
@@ -352,7 +366,13 @@ class FileIndex(collections.Mapping):
                     break
             else:
                 offsets.append((header_values, offsets_values))
-        return type(self)(filestream=self.filestream, index_keys=self.index_keys, offsets=offsets)
+        index = type(self)(
+            filestream=self.filestream,
+            index_keys=self.index_keys,
+            offsets=offsets,
+            filter_by_keys=query,
+        )
+        return index
 
     def first(self):
         with open(self.filestream.path) as file:
